#include "backend/bruteregalloc.hpp"
#include "backend/instr.hpp"
#include "common/regarch.hpp"

#include <iostream>

namespace backend {
namespace riscv {
using namespace RiscvReg;

const int INF = 0x3f3f3f3f;

BruteAllocator::BruteAllocator(Function* _func) : func(_func) {
    for (auto pi : func->stack_objects_offset_map_) {
        stack_offset[pi.first] = pi.second;
    }
    reg_alloc = std::vector<int>(RegCount, INF);
    for (int i = 0; i < func->arg_regs.size(); ++i) {
        reg_alloc[regs_arg[i]->id()] = func->arg_regs[i].id();
        reg_binding[func->arg_regs[i].id()] = regs_arg[i];
    }
    for (auto bb : *func->get_bbs()) {
        allocreg(bb);
    }
}

void BruteAllocator::allocreg(BasicBlock *bb) {
    if (!bb->get_instr()->size()) return;
    // bb->gen_asm_ori(std::cout);

    for (auto instr : *bb->get_instr()) {
        if (instr == bb->get_instr()->back()) break;
        allocreginstr(instr, bb);
    }

    auto instr = bb->get_instr()->back();
    auto instr_jump = dynamic_cast<RiscvInstr::Jump *>(instr);
    auto instr_branch = dynamic_cast<RiscvInstr::Branch *>(instr);
    
    if (!instr_jump && !instr_branch) allocreginstr(instr, bb);

    for (auto reg : bb->live_out) {
        // std::cout << "live_out " << reg.name() << std::endl;
        if (reg_binding.find(reg.id()) != reg_binding.end()) {
            StoreToStack(reg.id(), bb);
        }
    }

    if (instr_jump || instr_branch) allocreginstr(instr, bb);
}

void BruteAllocator::allocreginstr(Instr *instr, BasicBlock* bb) {
    // instr->gen_asm(std::cout);
    std::vector<Reg> def, use;
    for (auto & reg : instr->def_reg()) {
        def.push_back(reg.is_standard() ? reg : allocregreg(reg, false, instr->live_in, bb));
    }
    for (auto & reg : instr->use_reg()) {
        use.push_back(reg.is_standard() ? reg : allocregreg(reg, true, instr->live_in, bb));
    }
    bb->add_instr_after_alloc_(instr->tonative(def, use));
}

Reg BruteAllocator::allocregreg(Reg reg, bool read, std::set<Reg> live_in, BasicBlock* bb) {
    auto id = reg.id();
    if (reg_binding.find(id) != reg_binding.end()) {
        return reg_binding[id];
    }

    for (auto s_reg : regs_allocatable) {
        if (reg_alloc[s_reg->id()] == INF) {
            reg_alloc[s_reg->id()] = id;
            reg_binding[id] = s_reg;
            if (read) LoadFromStack(id, bb);
            return s_reg;
        }
    }

    for (auto s_reg : regs_allocatable) {
        if (!live_in.count(Reg(reg_alloc[s_reg->id()]))) {
            reg_binding.erase(reg_binding.find(reg_alloc[s_reg->id()]));
            reg_alloc[s_reg->id()] = id;
            reg_binding[id] = s_reg;
            if (read) LoadFromStack(id, bb);
            return s_reg;
        }
    }

    auto s_reg = regs_allocatable[std::rand() % RegAllocatableCount];
    StoreToStack(reg_alloc[s_reg->id()], bb);
    reg_binding.erase(reg_binding.find(reg_alloc[s_reg->id()]));
    reg_alloc[s_reg->id()] = id;
    reg_binding[id] = s_reg;
    if (read) LoadFromStack(id, bb);
    return s_reg;
}

void BruteAllocator::StoreToStack(int temp_reg_id, BasicBlock* bb) {
    assert(reg_binding.find(temp_reg_id) != reg_binding.end());
    if (stack_offset.find(temp_reg_id) == stack_offset.end()) {
        stack_offset[temp_reg_id] = func->stack_object_size_;
        func->stack_object_size_ += 4;
    }
    bb->add_instr_after_alloc_(new RiscvInstr::Store(reg_binding[temp_reg_id], SP, stack_offset[temp_reg_id]));
}

void BruteAllocator::LoadFromStack(int temp_reg_id, BasicBlock* bb) {
    assert(reg_binding.find(temp_reg_id) != reg_binding.end());
    assert(stack_offset.find(temp_reg_id) != stack_offset.end());
    bb->add_instr_after_alloc_(new RiscvInstr::Load(reg_binding[temp_reg_id], SP, stack_offset[temp_reg_id]));
}

} // namespace riscv

} // namespace backend