#include "backend/program.hpp"
#include "backend/instr.hpp"
#include "backend/bruteregalloc.hpp"
#include "backend/regAllocator.hpp"
#include "common/utils.hpp"
#include "common/regarch.hpp"
#include "middleend/ir.hpp"
#include <iostream>
#include <iomanip>

#define getValue(x) (x.type == Int ? x.iv : x.fv)

namespace backend{
namespace riscv{

using namespace RiscvReg;

std::vector<Function *> Program::functions() { return functions_; }

void Program::gen_asm(std::ostream &out) {
    gen_global_var_asm(out);
    out << "    .globl main\n";
    out << "    .text\n";

    /// compute used func
    std::vector<Function *> func_used;
    std::unordered_set<Function *> visited_funcs;
    std::queue<Function *> func_queue;
    for(auto &main_function: functions_){
        if(main_function->name() == "main") {
            func_used.push_back(main_function);
            func_queue.push(main_function);
            visited_funcs.insert(main_function);

            while (!func_queue.empty()) {
                Function *func_cur = func_queue.front();
                func_queue.pop();

                for (auto func_called : func_cur->call_funcs) {
                    for (auto &func : functions_) {
                        if (func->name() == func_called && visited_funcs.find(func) == visited_funcs.end()) {
                            func_used.push_back(func);
                            func_queue.push(func);
                            visited_funcs.insert(func);
                        }
                    }
                }
            }
            break;
        }
    }

    /// gen asm for used func
    for (auto &child : func_used) {
        child->gen_asm(out);
    }
}

bool BasicBlock::construct(ir::BasicBlock *ir_bb, BasicBlock *exit_bb, Function *func, Program *prog) {
    bool jump_to_another_block = false;
    for(auto ir_instr : *ir_bb->get_instructions()){
        // instrs_.push_back(new Instr(instr));
        if (auto ir_assign = dynamic_cast<ir::instruction::Assign*>(ir_instr)) {
            const Reg dst = Reg(ir_assign->getdst()->get_index(), false, ir_assign->getdst()->get_type().base_type == ScalarType::Int);
            const Reg src = Reg(ir_assign->getsrc()->get_index(), false, ir_assign->getsrc()->get_type().base_type == ScalarType::Int);
            instrs_.push_back(new RiscvInstr::Move(dst, src));
        } else if (auto ir_loadimm = dynamic_cast<ir::instruction::LoadImm4*>(ir_instr)) {
            if (ir_loadimm->getdst()->get_type().base_type == ScalarType::Int) {
                // std::cout << "# load int\n";
                const Reg dst = Reg(ir_loadimm->getdst()->get_index());
                const int imm = ir_loadimm->getimm().iv;
                instrs_.push_back(new RiscvInstr::LoadImm(dst, imm));
                func->constant_regs[dst] = imm;
            } else {
                // std::cout << "# load float\n";
                const Reg dst = Reg(func->reg_n++);
                float imm = ir_loadimm->getimm().fv;
                unsigned char *bytePointer = reinterpret_cast<unsigned char*>(&imm);
                unsigned int intValue = *reinterpret_cast<unsigned int*>(bytePointer);
                instrs_.push_back(new RiscvInstr::LoadImm(dst, intValue));
                const Reg dst_fp = Reg(ir_loadimm->getdst()->get_index(), false, false);
                instrs_.push_back(new RiscvInstr::Move(dst_fp, dst));
                func->constant_regs[dst] = intValue;
            }
            // add constant regs
        } else if (auto ir_unary = dynamic_cast<ir::instruction::Unary*>(ir_instr)) {
            // TODO to be check
            const Reg dst = Reg(ir_unary->getdst()->get_index(), false, ir_unary->getdst()->get_type().base_type == ScalarType::Int);
            const Reg src = Reg(ir_unary->getsrc()->get_index(), false, ir_unary->getsrc()->get_type().base_type == ScalarType::Int); 
            if (ir_unary->get_type() == UnaryOp::Add) {
                instrs_.push_back(new RiscvInstr::Move(dst, src));
            } else if (ir_unary->get_type() == UnaryOp::Sub) {
                if (dst.is_gp())
                    instrs_.push_back(new RiscvInstr::Unary(RiscvInstr::RvUnaryOp::NEG, dst, src));
                else
                    instrs_.push_back(new RiscvInstr::Unary(RiscvInstr::RvUnaryOp::FNEGS, dst, src));
            } else if (ir_unary->get_type() == UnaryOp::Not) {
                instrs_.push_back(new RiscvInstr::SLTIU(dst, src, 1));
            } else {
                throw std::runtime_error("not supported unary op");
            }
        } else if (auto ir_binary = dynamic_cast<ir::instruction::Binary*>(ir_instr)) {
            const Reg dst = Reg(ir_binary->getdst()->get_index(), false, ir_binary->getdst()->get_type().base_type == ScalarType::Int);
            const Reg lhs = Reg(ir_binary->getlhs()->get_index(), false, ir_binary->getlhs()->get_type().base_type == ScalarType::Int);
            const Reg rhs = Reg(ir_binary->getrhs()->get_index(), false, ir_binary->getrhs()->get_type().base_type == ScalarType::Int);
            if (ir_binary->get_type() == BinaryOp::Add) {
                // std::cout << "binary add dst " << dst << "\n";
                // std::cout << "binary lhs " << lhs << "\n"; 
                // std::cout << "binary rhs " << rhs << "\n";
                if (lhs.is_gp())
                    instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::ADD, dst, lhs, rhs));
                else
                    instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::FADDS, dst, lhs, rhs));
            } else if (ir_binary->get_type() == BinaryOp::Sub) {
                if (lhs.is_gp())
                    instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::SUB, dst, lhs, rhs));
                else
                    instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::FSUBS, dst, lhs, rhs));
            } else if (ir_binary->get_type() == BinaryOp::Mul) {
                if (lhs.is_gp())
                    instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::MUL, dst, lhs, rhs));
                else
                    instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::FMULS, dst, lhs, rhs));
            } else if (ir_binary->get_type() == BinaryOp::Div) {
                if (dst.is_gp())
                    instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::DIV, dst, lhs, rhs));
                else
                    instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::FDIVS, dst, lhs, rhs));
            } else if (ir_binary->get_type() == BinaryOp::Mod) {
                if (dst.is_gp())
                    instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::REM, dst, lhs, rhs));
                else
                    instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::FMODS, dst, lhs, rhs));
            } else if (ir_binary->get_type() == BinaryOp::Eq) {
                if (lhs.is_gp()) {
                    instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::SUB, dst, lhs, rhs));
                    instrs_.push_back(new RiscvInstr::Unary(RiscvInstr::RvUnaryOp::SEQZ, dst, dst));
                }
                else
                    instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::FEQ, dst, lhs, rhs));
            } else if (ir_binary->get_type() == BinaryOp::Neq) {
                if (lhs.is_gp()) {
                    instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::SUB, dst, lhs, rhs));
                    instrs_.push_back(new RiscvInstr::Unary(RiscvInstr::RvUnaryOp::SNEZ, dst, dst));
                }
                else {
                    instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::FEQ, dst, lhs, rhs));
                    instrs_.push_back(new RiscvInstr::Unary(RiscvInstr::RvUnaryOp::SNEZ, dst, dst));
                }
            } else if (ir_binary->get_type() == BinaryOp::Lt) {
                // instrs_.push_back(new RiscvInstr::Sext(lhs, lhs));
                // instrs_.push_back(new RiscvInstr::Sext(rhs, rhs));
                if (lhs.is_gp())
                    instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::SLT, dst, lhs, rhs));
                else
                    instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::FLT, dst, lhs, rhs));
            } else if (ir_binary->get_type() == BinaryOp::Gt) {
                if (lhs.is_gp())
                    instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::SLT, dst, rhs, lhs));
                else
                    instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::FLT, dst, rhs, lhs));
            } else if (ir_binary->get_type() == BinaryOp::Leq) {
                if (lhs.is_gp()) {
                    instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::SLT, dst, rhs, lhs));
                    instrs_.push_back(new RiscvInstr::Unary(RiscvInstr::RvUnaryOp::SEQZ, dst, dst));
                }
                else
                    instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::FLE, dst, lhs, rhs));
            } else if (ir_binary->get_type() == BinaryOp::Geq) {
                if (lhs.is_gp()) {
                    instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::SLT, dst, lhs, rhs));
                    instrs_.push_back(new RiscvInstr::Unary(RiscvInstr::RvUnaryOp::SEQZ, dst, dst));
                }
                else 
                    instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::FLE, dst, rhs, lhs));
            } else if (ir_binary->get_type() == BinaryOp::And) {
                instrs_.push_back(new RiscvInstr::Unary(RiscvInstr::RvUnaryOp::SNEZ, dst, lhs));
                instrs_.push_back(new RiscvInstr::Unary(RiscvInstr::RvUnaryOp::NEG, dst, dst));
                instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::AND, dst, dst, rhs));
                instrs_.push_back(new RiscvInstr::Unary(RiscvInstr::RvUnaryOp::SNEZ, dst, dst));
            } else if (ir_binary->get_type() == BinaryOp::Or) {
                instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::OR, dst, lhs, rhs));
                instrs_.push_back(new RiscvInstr::Unary(RiscvInstr::RvUnaryOp::SNEZ, dst, dst));
            } else if (ir_binary->get_type() == BinaryOp::Shr) {
                instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::SRL, dst, lhs, rhs));
            } else if (ir_binary->get_type() == BinaryOp::Shl) {
                instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::SLL, dst, lhs, rhs));
            } else {
                throw std::runtime_error("not supported binary op");
            }
        } else if (auto ir_label = dynamic_cast<ir::instruction::Label*>(ir_instr)) {
            instrs_.push_back(new RiscvInstr::RiscvLabel(ir_label->get_label()));
        } else if (auto ir_call = dynamic_cast<ir::instruction::Call*>(ir_instr)) {
            int gp_cnt = 0, fp_cnt = 0, stack_cnt = 0;
            // std::cout << "# call" << ir_call->getfunc()->get_name() << "\n";
            for (int i = 0; i < ir_call->get_srcs().size(); ++i) {
                bool is_gp_reg = ir_call->get_srcs()[i]->get_type().base_type == ScalarType::Int;
                if (ir_call->get_srcs()[i]->get_type().is_array())
                    is_gp_reg = true;
                const Reg arg = Reg(ir_call->get_srcs()[i]->get_index(), false, is_gp_reg);
                if (is_gp_reg) {
                    if (gp_cnt < RegArgCount) {
                        // std::cout << "# get gp cnt " << gp_cnt << "\n";
                        instrs_.push_back(new RiscvInstr::Move(*regs_arg[gp_cnt++], arg, true));
                    }
                    else {
                        gp_cnt++;
                        instrs_.push_back(new RiscvInstr::StoreDouble(arg, SP, 8 * (stack_cnt++), false, false, true));
                    }
                } 
                else {
                    if (fp_cnt < FpRegArgCount) {
                        // std::cout << "# get fp cnt " << fp_cnt << "\n";
                        instrs_.push_back(new RiscvInstr::Move(*fp_regs_arg[fp_cnt++], arg, true));
                    }
                    else {
                        fp_cnt++;
                        instrs_.push_back(new RiscvInstr::StoreDouble(arg, SP, 8 * (stack_cnt++), false, false, true));
                    }
                }
            }
            func->call_funcs.emplace_back(ir_call->getfunc()->get_name());
            instrs_.push_back(new RiscvInstr::Call(ir_call->getfunc()->get_name(), gp_cnt, fp_cnt));
            if(ir_call->getdst() != nullptr) {
                if (ir_call->getdst()->get_type().base_type == ScalarType::Int)
                    instrs_.push_back(new RiscvInstr::Move(Reg(ir_call->getdst()->get_index()), A0, true));
                else
                    instrs_.push_back(new RiscvInstr::Move(Reg(ir_call->getdst()->get_index(), false, false), FP10, true));
            }
        } else if (auto ir_branch = dynamic_cast<ir::instruction::Branch*>(ir_instr)) {
            instrs_.push_back(new RiscvInstr::Jump(prog->bb_map[ir_branch->get_bb()]));
            func->add_edge(this, prog->bb_map[ir_branch->get_bb()]);
            jump_to_another_block = true;
        } else if (auto ir_condbranch = dynamic_cast<ir::instruction::CondBranch*>(ir_instr)) {
            auto label = (ir_condbranch->get_type() == ir::instruction::CondBranch::IR_Instr_BEQ)
                    ?RiscvInstr::RvCompareOp::EQ : RiscvInstr::RvCompareOp::NE;
            auto false_bb = ir_condbranch->get_false_bb();
            auto true_bb = ir_condbranch->get_true_bb();
            instrs_.push_back(new RiscvInstr::Branch(label, Reg(ir_condbranch->getcond()->get_index()), prog->bb_map[false_bb]));
            instrs_.push_back(new RiscvInstr::Jump(prog->bb_map[true_bb]));
            func->add_edge(this, prog->bb_map[false_bb]);
            func->add_edge(this, prog->bb_map[true_bb]);
            jump_to_another_block = true;
        } else if (auto ir_return = dynamic_cast<ir::instruction::Return*>(ir_instr)) {
            // jump to exit_block
            if(ir_return->getvalue() != nullptr) {
                if (ir_return->getvalue()->get_type().base_type == ScalarType::Int) 
                    instrs_.push_back(new RiscvInstr::Move(A0, Reg(ir_return->getvalue()->get_index())));
                else
                    instrs_.push_back(new RiscvInstr::Move(FP10, Reg(ir_return->getvalue()->get_index(), false, false)));
            }
            instrs_.push_back(new RiscvInstr::Jump(exit_bb));
            func->add_edge(this, exit_bb);
            jump_to_another_block = true;
        } else if (auto ir_alloca = dynamic_cast<ir::instruction::Alloca*>(ir_instr)) {
            // Do nothing.
            if(func->stack_objects_offset_map_[ir_alloca->getaddr()->get_index()] > 2047 || func->stack_objects_offset_map_[ir_alloca->getaddr()->get_index()] < -2048){
                Reg new_reg = Reg(ir_alloca->getaddr()->get_index());
                Reg new_reg_to_save_imm = Reg(func->reg_n++);
                instrs_.push_back(new RiscvInstr::LoadImm(new_reg_to_save_imm, func->stack_objects_offset_map_[ir_alloca->getaddr()->get_index()]));
                instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::ADD, new_reg, SP, new_reg_to_save_imm));
            }
            else{
                // std::cout << "# 【ir_alloca】 offset is " << func->stack_objects_offset_map_[ir_alloca->getaddr()->get_index()] << "\n";
                instrs_.push_back(new RiscvInstr::ADDImm(Reg(ir_alloca->getaddr()->get_index()), SP, func->stack_objects_offset_map_[ir_alloca->getaddr()->get_index()]));
            }
        } else if (auto ir_store = dynamic_cast<ir::instruction::Store*>(ir_instr)) {
            const Reg addr = Reg(ir_store->getaddr()->get_index());
            const Reg src = Reg(ir_store->getsrc()->get_index(), false, ir_store->getsrc()->get_type().base_type == ScalarType::Int);
            //instrs_.push_back(new RiscvInstr::Store(src, addr, ir_store->getoffset()));
            if(ir_store->getoffset() > 2047 || ir_store->getoffset() < -2048){
                Reg new_reg_to_save_imm = Reg(func->reg_n++);
                instrs_.push_back(new RiscvInstr::LoadImm(new_reg_to_save_imm, ir_store->getoffset()));
                instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::ADD, new_reg_to_save_imm, new_reg_to_save_imm, addr));
                instrs_.push_back(new RiscvInstr::Store(src, new_reg_to_save_imm, 0));
            }
            else{
                instrs_.push_back(new RiscvInstr::Store(src, addr, ir_store->getoffset()));
            }
        } else if (auto ir_load = dynamic_cast<ir::instruction::Load*>(ir_instr)) {
            const Reg dst = Reg(ir_load->getdst()->get_index(), false, ir_load->getdst()->get_type().base_type == ScalarType::Int);
            const Reg addr = Reg(ir_load->getaddr()->get_index());
            //instrs_.push_back(new RiscvInstr::Load(dst, addr, ir_load->getoffset()));

            if(ir_load->getoffset() > 2047 || ir_load->getoffset() < -2048){
                Reg new_reg_to_save_imm = Reg(func->reg_n++);
                instrs_.push_back(new RiscvInstr::LoadImm(new_reg_to_save_imm, ir_load->getoffset()));
                instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::ADD, addr, new_reg_to_save_imm, addr));
                instrs_.push_back(new RiscvInstr::Load(dst, addr, 0));
            }
            else{
                instrs_.push_back(new RiscvInstr::Load(dst, addr, ir_load->getoffset()));
            }

        } else if (auto ir_loadaddr = dynamic_cast<ir::instruction::LoadAddr*>(ir_instr)) {
            const Reg dst = Reg(ir_loadaddr->get_addr()->get_index());
            instrs_.push_back(new RiscvInstr::LoadLabelAddr(dst, ir_loadaddr->get_name()));
        } else if (auto ir_elementptr = dynamic_cast<ir::instruction::ElementPtr*>(ir_instr)) {
            auto ir_srcs = *ir_elementptr->get_src();
            // 每一维的下标：
            auto indices = std::vector<Temp*>(ir_srcs.begin() + 1, ir_srcs.end());
            // 每一维的大小
            auto dims = ir_elementptr->get_dst()->get_type().dims;

            Reg base = Reg(ir_elementptr->get_base()->get_index());

            int dim_size = 4;
            // 从最后一维往前
            for(int i = dims.size() - 1; i >= 0; dim_size *= dims[i], i--){
                if(i >= indices.size())continue;
                //std::cout << func->reg_n << std::endl;
                const Reg size = (i == 0) ? Reg(ir_elementptr->get_dst()->get_index()) : Reg(func->reg_n++);
                // std::cout << ir_elementptr->to_str() << std::endl;
                // std::cout << func->reg_n << std::endl;
                // std::cout << size << std::endl;
                instrs_.push_back(new RiscvInstr::LoadImm(size, dim_size));
                const Reg index = Reg(indices[i]->get_index());

                instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::MUL, size, size, index));
                instrs_.push_back(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::ADD, size, size, base));
                base = size;
            }
        } else if (auto ir_phi = dynamic_cast<ir::instruction::Phi*>(ir_instr)) {
            // do nothing here
            assert(true);
        } else if (auto ir_cast = dynamic_cast<ir::instruction::Cast*>(ir_instr)) {
            const Reg dst = Reg(ir_cast->getdst()->get_index(), false, ir_cast->getdst()->get_type().base_type == ScalarType::Int);
            const Reg src = Reg(ir_cast->getsrc()->get_index(), false, ir_cast->getsrc()->get_type().base_type == ScalarType::Int);
            instrs_.push_back(new RiscvInstr::Convert(dst, src));
        } else {
            assert(false);
        }
    }
    // std::cout << "----debug----\n";
    // for (auto i: instrs_) 
    //     i->gen_asm(std::cout);
    return jump_to_another_block;
}

void BasicBlock::gen_asm_ori(std::ostream &out){
    if(label_used) out << name_ << ":\n";
    for (auto instr : instrs_) instr->gen_asm(out);
}

void BasicBlock::gen_asm(std::ostream &out, int stack_object_size_){
    if(label_used) out << name_ << ":\n";
    // complete caller saved load and store
    std::set<Reg> temp;
    // out << "---" << name_ << ":\n";
    // for(auto succ_: succ){
    //     out << "-------" << succ_->get_name() << "\n";
    //     for(auto reg: succ_->live_use){
    //         out << "-:-:-:" << reg << "\n";
    //     }
    // }
    for (auto instr_it = instrs_after_alloc_.begin(); instr_it != instrs_after_alloc_.end(); instr_it++){
    //     if (auto ir_call = dynamic_cast<RiscvInstr::Store*>(*instr_it)) {
    //         if(ir_call->is_for_func) {
    //             auto dst_reg = ir_call->use_reg()[0];
    //             if(this->live_out.find(dst_reg) == this->live_out.end()) {
    //                 bool can_be_used_later = false;
    //                 for(auto itt = instr_it; itt != instrs_after_alloc_.end(); itt++){
    //                     if(itt == instr_it) continue;
    //                     if(auto ir_call = dynamic_cast<RiscvInstr::Call*>(*itt)) continue;
    //                     if (auto ir_mv = dynamic_cast<RiscvInstr::Move*>(*itt)) {
    //                         if(ir_mv->pass_arg){
    //                             continue;
    //                         }
    //                     }
    //                     if (auto ir_store = dynamic_cast<RiscvInstr::Store*>(*itt)) {
    //                         if(ir_store->is_for_func){
    //                             if(ir_store->use_reg()[1].id() != dst_reg.id()){
    //                                 continue;
    //                             } else {
    //                                 can_be_used_later = true;
    //                                 break;
    //                             }
    //                         }
    //                         if(ir_store->arg){
    //                             if(ir_store->use_reg()[1].id() != dst_reg.id()){
    //                                 continue;
    //                             }
    //                         }
    //                     }
                        
    //                     auto inst = *itt;
    //                     auto use_reg = inst->use_reg();
    //                     auto def_reg = inst->def_reg();
    //                     if(std::find(use_reg.begin(), use_reg.end(), dst_reg) != use_reg.end()) {
    //                         can_be_used_later = true;
    //                         break;
    //                     }
    //                     if(std::find(def_reg.begin(), def_reg.end(), dst_reg) != def_reg.end()) {
    //                         if(auto ir_load = dynamic_cast<RiscvInstr::Load*>(*itt)){
    //                             if(ir_load->is_for_func){
    //                                 continue;
    //                             }
    //                             if(ir_load->arg){
    //                                 continue;
    //                             }
    //                         }
    //                         can_be_used_later = false;
    //                         break;
    //                     }
    //                 }
    //                 if(!can_be_used_later) {
    //                     // std::cout << "# delete:" << std::endl; 
    //                     // (*instr_it)->gen_asm(out);
    //                     // std::cout << "# delete" << std::endl; 
    //                     continue;
    //                 }
    //             }
    //             std::cout << "# store caller_saved reg:" << std::endl; 
    //         }
    //     }
    //     else if (auto ir_call = dynamic_cast<RiscvInstr::Load*>(*instr_it)) {
    //         if(ir_call->is_for_func) {
    //             auto dst_reg = ir_call->def_reg()[0];
    //             if(this->live_out.find(dst_reg) == this->live_out.end()) {
    //                 bool can_be_used_later = false;
    //                 for(auto itt = instr_it; itt != instrs_after_alloc_.end(); itt++){
    //                     if(itt == instr_it) continue;
    //                     if(auto ir_call = dynamic_cast<RiscvInstr::Call*>(*itt)) continue;
    //                     if (auto ir_mv = dynamic_cast<RiscvInstr::Move*>(*itt)) {
    //                         if(ir_mv->pass_arg){
    //                             continue;
    //                         }
    //                     }
    //                     if (auto ir_store = dynamic_cast<RiscvInstr::Store*>(*itt)) {
    //                         if(ir_store->is_for_func){
    //                             if(ir_store->use_reg()[1].id() != dst_reg.id()){
    //                                 continue;
    //                             } else {
    //                                 can_be_used_later = true;
    //                                 break;
    //                             }
    //                         }
    //                     }
    //                     auto inst = *itt;
    //                     auto use_reg = inst->use_reg();
    //                     auto def_reg = inst->def_reg();
    //                     if(std::find(use_reg.begin(), use_reg.end(), dst_reg) != use_reg.end()) {
    //                         can_be_used_later = true;
    //                         break;
    //                     }
    //                     if(std::find(def_reg.begin(), def_reg.end(), dst_reg) != def_reg.end()) {
    //                         if(auto ir_load = dynamic_cast<RiscvInstr::Load*>(*itt)){
    //                             if(ir_load->is_for_func){
    //                                 continue;
    //                             }
    //                         }
    //                         can_be_used_later = false;
    //                         break;
    //                     }
    //                 }
    //                 if(!can_be_used_later) continue;
    //             } 
    //             std::cout << "# load caller_saved reg:" << std::endl; 
    //         }
    //     }
        (*instr_it)->gen_asm(out);
        // if (auto ir_call = dynamic_cast<RiscvInstr::Call*>(*instr_it)) {
        //     for(auto r: temp){
        //         auto new_instr = new RiscvInstr::Load(r, SP, stack_object_size_ + 4 * (caller_to_offset[r.id()] + RegCalleeSavedCount));
        //         new_instr->gen_asm(out);
        //     }
        // }
    }
}

// static void find_caller_saved_reg(BasicBlock* bb, std::vector<Reg>* caller_saved, int stack_object_size, int callee_reg_size){
//     // std::cout << bb->get_name() << " old version:" << std::endl;
//     // for (auto instr_it = bb->get_instr_after_alloc()->begin(); instr_it != bb->get_instr_after_alloc()->end(); instr_it++){
//     //     (*instr_it)->gen_asm(std::cout);
//     // }
//     for (auto instr_it = bb->get_instr_after_alloc()->begin(); instr_it != bb->get_instr_after_alloc()->end(); instr_it++){
//         //std::cout << "aaa:";
//         //(*instr_it)->gen_asm(std::cout);
//         if (auto ir_call = dynamic_cast<RiscvInstr::Store*>(*instr_it)) {
//             if(ir_call->is_for_func) {
//                 auto dst_reg = ir_call->use_reg()[0];
//                 if(bb->live_out.find(dst_reg) == bb->live_out.end()) {
//                     bool can_be_used_later = false;
//                     for(auto itt = instr_it; itt != bb->get_instr_after_alloc()->end(); itt++){
//                         if(itt == instr_it) continue;
//                         if(auto ir_call = dynamic_cast<RiscvInstr::Call*>(*itt)) continue;
//                         if (auto ir_mv = dynamic_cast<RiscvInstr::Move*>(*itt)) {
//                             if(ir_mv->pass_arg){
//                                 continue;
//                             }
//                         }
//                         if (auto ir_store = dynamic_cast<RiscvInstr::Store*>(*itt)) {
//                             if(ir_store->is_for_func){
//                                 if(ir_store->use_reg()[1].id() != dst_reg.id()){
//                                     continue;
//                                 } else {
//                                     can_be_used_later = true;
//                                     break;
//                                 }
//                             }
//                             if(ir_store->arg){
//                                 if(ir_store->use_reg()[1].id() != dst_reg.id()){
//                                     continue;
//                                 }
//                             }
//                         }                   
//                         auto inst = *itt;
//                         auto use_reg = inst->use_reg();
//                         auto def_reg = inst->def_reg();
//                         if(std::find(use_reg.begin(), use_reg.end(), dst_reg) != use_reg.end()) {
//                             can_be_used_later = true;
//                             break;
//                         }
//                         if(std::find(def_reg.begin(), def_reg.end(), dst_reg) != def_reg.end()) {
//                             if(auto ir_load = dynamic_cast<RiscvInstr::Load*>(*itt)){
//                                 if(ir_load->is_for_func){
//                                     continue;
//                                 }
//                                 if(ir_load->arg){
//                                     continue;
//                                 }
//                             }
//                             can_be_used_later = false;
//                             break;
//                         }
//                     }
//                     if(!can_be_used_later) {
//                         //std::cout << "erase!!!";
//                         //ir_call->gen_asm(std::cout);
//                         if(ir_call->offsetZero()){
//                             instr_it = bb->get_instr_after_alloc()->erase(instr_it);
//                             instr_it--;
//                             instr_it = bb->get_instr_after_alloc()->erase(instr_it);
//                             instr_it--;
//                             instr_it = bb->get_instr_after_alloc()->erase(instr_it);
//                             instr_it--;
//                         }else {
//                             instr_it = bb->get_instr_after_alloc()->erase(instr_it);
//                             instr_it--;
//                         }
//                         continue;
//                     } else {
//                         auto ir = std::find(caller_saved->begin(), caller_saved->end(), dst_reg);
//                         int offset_to_change = stack_object_size + callee_reg_size;
//                         if(ir == caller_saved->end()){
//                             offset_to_change += caller_saved->size() * 4;
//                             caller_saved->push_back(dst_reg);
//                         } else {
//                             size_t irIndex = std::distance(caller_saved->begin(), ir);
//                             offset_to_change += irIndex * 4;
//                         }
//                         if(ir_call->offsetZero()){
//                             // 调用ir_call的上一条的上一条指令的changeOffset函数
//                             auto prev = std::prev(std::prev(instr_it));
//                             if(auto ld = dynamic_cast<RiscvInstr::LoadImm*>(*prev)){
//                                 ld->changeVal(offset_to_change);
//                             } else {
//                                 assert(false), "error changing offset";
//                             }
//                         }else{
//                             ir_call->changeOffset(offset_to_change);
//                         }
//                     }
//                 } else {
//                     auto ir = std::find(caller_saved->begin(), caller_saved->end(), dst_reg);
//                     int offset_to_change = stack_object_size + callee_reg_size;
//                     if(ir == caller_saved->end()){
//                         offset_to_change += caller_saved->size() * 4;
//                         caller_saved->push_back(dst_reg);
//                     } else {
//                         size_t irIndex = std::distance(caller_saved->begin(), ir);
//                         offset_to_change += irIndex * 4;
//                     }
//                     if(ir_call->offsetZero()){
//                         // 调用ir_call的上一条的上一条指令的changeOffset函数
//                         auto prev = std::prev(std::prev(instr_it));
//                         if(auto ld = dynamic_cast<RiscvInstr::LoadImm*>(*prev)){
//                             ld->changeVal(offset_to_change);
//                         } else {
//                             assert(false), "error changing offset";
//                         }
//                     }else{
//                         ir_call->changeOffset(offset_to_change);
//                     }
//                 }
//             }
//         }
//         else if (auto ir_call = dynamic_cast<RiscvInstr::Load*>(*instr_it)) {
//             //ir_call->gen_asm(std::cout);
//             if(ir_call->is_for_func) {
//                 auto dst_reg = ir_call->def_reg()[0];
//                 //ir_call->gen_asm(std::cout);
//                 if(bb->live_out.find(dst_reg) == bb->live_out.end()) {
//                     bool can_be_used_later = false;
//                     for(auto itt = instr_it; itt != bb->get_instr_after_alloc()->end(); itt++){
//                         if(itt == instr_it) continue;
//                         if(auto ir_call = dynamic_cast<RiscvInstr::Call*>(*itt)) continue;
//                         if (auto ir_mv = dynamic_cast<RiscvInstr::Move*>(*itt)) {
//                             if(ir_mv->pass_arg){
//                                 continue;
//                             }
//                         }
//                         if (auto ir_store = dynamic_cast<RiscvInstr::Store*>(*itt)) {
//                             if(ir_store->is_for_func){
//                                 if(ir_store->use_reg()[1].id() != dst_reg.id()){
//                                     continue;
//                                 } else {
//                                     //ir_store->gen_asm(std::cout);
//                                     can_be_used_later = true;
//                                     break;
//                                 }
//                             }
//                             if(ir_store->arg){
//                                 if(ir_store->use_reg()[1].id() != dst_reg.id()){
//                                     continue;
//                                 }
//                             }
//                         }
                        
//                         auto inst = *itt;
//                         auto use_reg = inst->use_reg();
//                         auto def_reg = inst->def_reg();
//                         if(std::find(use_reg.begin(), use_reg.end(), dst_reg) != use_reg.end()) {
//                             //inst->gen_asm(std::cout);
//                             can_be_used_later = true;
//                             break;
//                         }
//                         if(std::find(def_reg.begin(), def_reg.end(), dst_reg) != def_reg.end()) {
//                             if(auto ir_load = dynamic_cast<RiscvInstr::Load*>(*itt)){
//                                 if(ir_load->is_for_func){
//                                     continue;
//                                 }
//                                 if(ir_load->arg){
//                                     continue;
//                                 }
//                             }
//                             can_be_used_later = false;
//                             break;
//                         }
//                     }
//                     if(!can_be_used_later) {
//                         //std::cout << "erase!!!";
//                         //ir_call->gen_asm(std::cout);
//                         if(ir_call->offsetZero()){
//                             instr_it = bb->get_instr_after_alloc()->erase(instr_it);
//                             instr_it--;
//                             instr_it = bb->get_instr_after_alloc()->erase(instr_it);
//                             instr_it--;
//                             instr_it = bb->get_instr_after_alloc()->erase(instr_it);
//                             instr_it--;
//                         }else {
//                             instr_it = bb->get_instr_after_alloc()->erase(instr_it);
//                             instr_it--;
//                         }
//                         continue;
//                     } else {
//                         auto ir = std::find(caller_saved->begin(), caller_saved->end(), dst_reg);
//                         int offset_to_change = stack_object_size + callee_reg_size;
//                         size_t irIndex = std::distance(caller_saved->begin(), ir);
//                         offset_to_change += irIndex * 4;
//                         if(ir_call->offsetZero()){
//                             // 调用ir_call的上一条的上一条指令的changeOffset函数
//                             auto prev = std::prev(std::prev(instr_it));
//                             if(auto ld = dynamic_cast<RiscvInstr::LoadImm*>(*prev)){
//                                 ld->changeVal(offset_to_change);
//                             } else {
//                                 assert(false), "error changing offset";
//                             }
//                         }else{
//                             ir_call->changeOffset(offset_to_change);
//                         }
//                     }
//                 } else {
//                     auto ir = std::find(caller_saved->begin(), caller_saved->end(), dst_reg);
//                     int offset_to_change = stack_object_size + callee_reg_size;
//                     size_t irIndex = std::distance(caller_saved->begin(), ir);
//                     offset_to_change += irIndex * 4;
//                     if(ir_call->offsetZero()){
//                         // 调用ir_call的上一条的上一条指令的changeOffset函数
//                         auto prev = std::prev(std::prev(instr_it));
//                         if(auto ld = dynamic_cast<RiscvInstr::LoadImm*>(*prev)){
//                             ld->changeVal(offset_to_change);
//                         } else {
//                             assert(false), "error changing offset";
//                         }
//                     }else{
//                         ir_call->changeOffset(offset_to_change);
//                     }
//                 }
//             }
//         }
//     }
//     // std::cout << bb->get_name() << " new version:" << std::endl;
//     // for (auto instr_it = bb->get_instr_after_alloc()->begin(); instr_it != bb->get_instr_after_alloc()->end(); instr_it++){
//     //     (*instr_it)->gen_asm(std::cout);
//     // }
//     return;
// }

Function::Function(Program *prog, ir::Function *ir_func): name_(ir_func->get_name()), reg_n(ir_func->get_temp_used()), stack_object_size_(0), stack_spill_size_(0) {
    this->call_funcs.clear();
    BasicBlock* entry_bb = new BasicBlock(".entry_" + name_);
    entry_bb->set_label_used(true);
    int gp_cnt = 0, fp_cnt = 0, stack_cnt = 0;
    for (int i = 0; i < ir_func->get_arg_num(); ++i) {
        bool is_gp = (*ir_func->get_arg_temp())[i]->get_type().base_type == ScalarType::Int;
        if ((*ir_func->get_arg_temp())[i]->get_type().is_array())
            is_gp = true;
        if (is_gp) {
            if (gp_cnt < RegArgCount) {
                // std::cout << "# " << name_ << " entry gp cnt " << gp_cnt << "\n";
                entry_bb->get_instr()->push_back(new RiscvInstr::Move(Reg((*ir_func->get_arg_temp())[i]->get_index()), *regs_arg[gp_cnt++]));
            } else {
                entry_bb->get_instr()->push_back(new RiscvInstr::LoadDouble(Reg((*ir_func->get_arg_temp())[i]->get_index()), FP, (stack_cnt++) * 8));
            }
        }
        else {
            if (fp_cnt < FpRegArgCount) {
                // std::cout << "# " << name_ << " entry fp cnt " << fp_cnt << "\n";
                entry_bb->get_instr()->push_back(new RiscvInstr::Move(Reg((*ir_func->get_arg_temp())[i]->get_index(), false, false), *fp_regs_arg[fp_cnt++]));
            } else {
                entry_bb->get_instr()->push_back(new RiscvInstr::LoadDouble(Reg((*ir_func->get_arg_temp())[i]->get_index(), false, false), FP, (stack_cnt++) * 8));
            }
        }
    }
    bbs.push_back(std::move(entry_bb));
    BasicBlock* exit_bb = new BasicBlock(".exit_" + name_);
    exit_bb->set_label_used(true);

    entry_bb->get_instr()->push_front(new RiscvInstr::LoadImm(Reg(this->reg_n++), 0));

    exit_bb->get_instr()->push_back(new RiscvInstr::Return(ir_func->has_return_value(), ir_func->get_ret_type().base_type != ScalarType::Float));
    exit_bb->get_instr()->push_front(new RiscvInstr::LoadImm(Reg(this->reg_n++), 0));
    for(auto bb : *ir_func->get_basic_blocks()) {
        BasicBlock* new_bb = new BasicBlock(".L" + std::to_string(prog->bb_num++));
        new_bb->set_label_used(true); // TODO 灵车起飞
        prog->bb_map[bb] = new_bb;
        bbs.push_back(std::move(new_bb));
        //std::cout << new_bb->get_name() << ":" << &(*new_bb) << " " << &(*bbs.back()) << " " << &(*prog->bb_map[bb]) << std::endl;
    }

    auto prev_entry = prog->bb_map[ir_func->get_basic_blocks()->front()];
    add_edge(entry_bb, prev_entry);

    bool is_call_instr_exist_ = false;
    int max_call_args_size = 0;
    for(auto bb : *ir_func->get_basic_blocks()) {
        // calculate for stack object size
        for (auto ir_instr : *bb->get_instructions()) {
            if (auto ir_alloca = dynamic_cast<ir::instruction::Alloca*>(ir_instr)) {
                // std::cout << "# 【ir_alloca】 size is " << ir_alloca->get_size() << "\n";
                stack_objects_offset_map_[ir_alloca->getaddr()->get_index()] = stack_object_size_;
                stack_object_size_ += ir_alloca->get_size();
                // std::cout << "alloca instr exist " << ir_alloca->get_size() << std::endl;
            } else if (auto ir_call = dynamic_cast<ir::instruction::Call*>(ir_instr)) {
                is_call_instr_exist_ = true;
                max_call_args_size = ((ir_call->get_srcs()).size() > max_call_args_size) ? (ir_call->get_srcs()).size() : max_call_args_size;
                // std::cout << "call instr exist" << std::endl;
            }
        }
    }
    if(max_call_args_size > RegArgCount) {
        stack_object_size_ += 8 * (max_call_args_size - RegArgCount);
        for(auto it = stack_objects_offset_map_.begin(); it != stack_objects_offset_map_.end(); it++){
            it->second += 8 * (max_call_args_size - RegArgCount);
        }
    }
    stack_reg_size_ = RegCalleeSavedCount;
    if (is_call_instr_exist_) stack_reg_size_ += RegCallerSavedCount * 8;
    for(auto it = (*ir_func->get_basic_blocks()).begin(); it != (*ir_func->get_basic_blocks()).end(); it++) {
        auto bb = *it;
        // construct basic block instructions
        if(!prog->bb_map[bb]->construct(bb, exit_bb, this, prog) && std::next(it) != (*ir_func->get_basic_blocks()).end()){
            // add next block to this block's succ
            this->add_edge(prog->bb_map[bb], prog->bb_map[(*std::next(it))]);
        }
    }
    bbs.push_back(std::move(exit_bb));

    // deal with phi instructions here
    struct PhiMove{
        BasicBlock* src_block;
        BasicBlock* dst_block;
        Reg dst,src,mid;
    };
    std::map<int, Reg> dst_to_mid;
    std::vector<PhiMove> phi_moves;
    for(auto bb : *ir_func->get_basic_blocks()) {
        for(auto ir_instr : *bb->get_instructions()) {
            if(auto ir_phi = dynamic_cast<ir::instruction::Phi*>(ir_instr)) {
                for(int i = 0; i < ir_phi->getsize(); i++) {
                    auto ori_bb = (*ir_phi->getbbs())[i];
                    auto phi_src = Reg((*ir_phi->getvalues())[i]->get_index());
                    auto phi_mid = Reg(this->reg_n++);
                    auto phi_dst = Reg(ir_phi->getdst()->get_index());
                    if(dst_to_mid.find(ir_phi->getdst()->get_index()) != dst_to_mid.end()){
                        phi_mid = dst_to_mid.at(ir_phi->getdst()->get_index());
                        this->reg_n--;
                    } else {
                        dst_to_mid[ir_phi->getdst()->get_index()] = phi_mid;
                    }
                    //std::cout << "phi:" << prog->bb_map[ori_bb]->get_name() << ":" << &(*prog->bb_map[ori_bb]) << std::endl;
                    phi_moves.push_back({std::move(prog->bb_map[ori_bb]), std::move(prog->bb_map[bb]), phi_dst, phi_src, phi_mid});
                }
            }
        }
    }
    for(auto phi_move: phi_moves) {
        // std::cout << "# phi: " << phi_move.src_block->get_name() << " to " << phi_move.dst_block->get_name() << " " << phi_move.src << " to " << phi_move.dst << std::endl;
        //std::cout << "phi:" << phi_move.block->get_name() << ":" << &(*phi_move.block) << std::endl;
        auto instrs = phi_move.src_block->get_instr();
        auto new_instr = new RiscvInstr::Move(phi_move.dst, phi_move.src);
        for(auto it = instrs->begin(); it != instrs->end();){
            auto inst = *it;
            if(auto ir_jmp = dynamic_cast<RiscvInstr::Jump*>(inst)){
                if(ir_jmp->get_target()->get_name() == phi_move.dst_block->get_name()){
                    it = instrs->insert(it, new_instr);
                    it++;
                }
            }
            else if(auto ir_jmp = dynamic_cast<RiscvInstr::Branch*>(inst)){
                if(ir_jmp->get_target()->get_name() == phi_move.dst_block->get_name()){
                    it = instrs->insert(it, new_instr);
                    it++;
                }
            } 
            it++;
        }
        //phi_move.src_block->insert_before_jmp(phi_move.dst_block, new RiscvInstr::Move(phi_move.mid, phi_move.src));
        //phi_move.dst_block->add_instr_before_alloc(new RiscvInstr::Move(phi_move.dst, phi_move.mid), phi_move.dst_block->get_instr()->begin());
        //phi_move.block->change(instrs);
    }

    // livenessAnalysis(true);
    // gen_asm_ori(std::cout);

    // for (auto bb : bbs) {
    //     // print livein and liveout
    //     std::cout << "bb: " << bb->get_name() << std::endl;
    //     std::cout << "livein: ";
    //     for (auto &e : bb->live_in) std::cout << e.name() << " ";
    //     std::cout << std::endl;
    //     std::cout << "liveout: ";
    //     for (auto &e : bb->live_out) std::cout << e.name() << " ";
    //     std::cout << std::endl;
    // }
}

void Function::gen_asm(std::ostream &out) {
    //for (int i = 0; i < this->callee_saved_used.size(); i++) {
    //    std::cout << this->callee_saved_used[i] << std::endl;
    //}
    /// compute used bbs
    std::vector<BasicBlock *> bbs_used;
    std::vector<Reg>* caller_saved_reg_map = new std::vector<Reg>;
    // find_caller_saved_reg(bbs[0], caller_saved_reg_map, stack_object_size_, callee_saved_used.size() * 4);
    bbs_used.push_back(bbs[0]);
    std::unordered_set<BasicBlock *> visited_bbs;

    std::queue<BasicBlock *> bb_queue;
    bb_queue.push(bbs[0]);
    visited_bbs.insert(bbs[0]);

    while (!bb_queue.empty()) {
        BasicBlock *bb_cur = bb_queue.front();
        bb_queue.pop();

        for (auto bb_used : bb_cur->succ) {
            if (visited_bbs.find(bb_used) == visited_bbs.end()) {
                // find_caller_saved_reg(bb_used, caller_saved_reg_map, stack_object_size_, callee_saved_used.size() * 4);
                bbs_used.push_back(std::move(bb_used));
                bb_queue.push(bb_used);
                visited_bbs.insert(bb_used);
            }
        }
    }

    auto front_instr_front = bbs[0]->first_instr();
    auto front_dst_reg = front_instr_front->def_reg()[0];
    auto end_instr_front = bbs[bbs.size() - 1]->first_instr();
    auto end_dst_reg = end_instr_front->def_reg()[0];
    stack_reg_size_ = callee_saved_used.size() * 8 + caller_saved_reg_map->size() * 8;
    // std::cout << "# " << this->name_ << ":::" << stack_object_size_ << ":::" << stack_reg_size_ << ":::" << stack_spill_size_ << std::endl;
    if(auto loadimm = dynamic_cast<RiscvInstr::LoadImm*>(front_instr_front)){
        auto size = (stack_object_size_ + stack_reg_size_ + stack_spill_size_ + 7)/8 * 8;
        loadimm->update_imm(size);
    } else {
        assert(false);
    }
    if(auto loadimm = dynamic_cast<RiscvInstr::LoadImm*>(end_instr_front)){
        auto size = (stack_object_size_ + stack_reg_size_ + stack_spill_size_ + 7)/8 * 8;
        loadimm->update_imm(size);
    } else {
        assert(false);
    }
    bbs[0]->add_instr_after_alloc_second(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::ADD, FP, SP, front_dst_reg));
    bbs[0]->add_instr_after_alloc_second(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::SUB, SP, SP, front_dst_reg));
    for (int i = 0; i < this->callee_saved_used.size(); i++){
        bbs[0]->add_instr_after_alloc_front(new RiscvInstr::StoreDouble(this->callee_saved_used[i], SP, i * 8 - stack_reg_size_ - stack_spill_size_));
    }
    
    for (int i = 0; i < this->callee_saved_used.size(); i++)
        bbs[bbs.size() - 1]->add_instr_after_alloc_second(new RiscvInstr::LoadDouble(this->callee_saved_used[i], SP, i * 8 - stack_reg_size_ - stack_spill_size_));
    bbs[bbs.size() - 1]->add_instr_after_alloc_second(new RiscvInstr::Binary(RiscvInstr::RvBinaryOp::ADD, SP, SP, end_dst_reg));
    out << name_ << ":\n";

    /// gen asm for used bbs
    for (int i = 0; i < bbs_used.size(); ++i) bbs_used[i]->gen_asm(out, this->stack_object_size_);
}

void Function::gen_asm_ori(std::ostream &out) {
    for (int i = 0; i < RegCalleeSavedCount; i++)
        bbs[0]->get_instr()->push_front(new RiscvInstr::Store(regs_calleesaved[i], SP, i * 8 + stack_object_size_));
    bbs[0]->get_instr()->push_front(new RiscvInstr::ADDImm(FP, SP, stack_object_size_ + stack_reg_size_));
    bbs[0]->get_instr()->push_front(new RiscvInstr::ADDImm(SP, SP, -(stack_object_size_ + stack_reg_size_ + stack_spill_size_)));
    bbs[bbs.size() - 1]->get_instr()->push_front(new RiscvInstr::ADDImm(SP, SP, (stack_object_size_ + stack_reg_size_ + stack_spill_size_)));
    for (int i = 0; i < RegCalleeSavedCount; i++)
        bbs[bbs.size() - 1]->get_instr()->push_front(new RiscvInstr::Load(regs_calleesaved[i], SP, i * 8 + stack_object_size_));
    out << name_ << ":\n";
    for (int i = 0; i < bbs.size(); ++i) bbs[i]->gen_asm_ori(out);
}

Program::Program(middleend::ir::Module* module) {
    bb_num = 0;
    auto funcs = module -> get_functions();

    for (auto func : *module->get_functions()) {
        functions_.push_back(new Function(this, func));
        //std::cout << "end:" << functions_.back()->name() << std::endl;
        //std::cout << (functions_.back())->reg_n << std::endl;
        //std::cout << "start alloc reg:" << functions_.back()->name() << std::endl;
        ColoringRegAllocator color_reg_alloc;
        //auto allocator = BruteAllocator(functions_.back());
        color_reg_alloc.alloc_reg(functions_.back(), true);
        functions_.back()->callee_saved_used.clear();
        functions_.back()->callee_saved_used.emplace_back(RA);
        functions_.back()->callee_saved_used.emplace_back(FP);
        for(auto used: color_reg_alloc.used_reg) {
            if(regs_use[used.id()] == callee_save) {
                if(!(used == RA) && !(used == FP)){
                    functions_.back()->callee_saved_used.emplace_back(used);
                }
            }
        }
        // std::cout << "# ----\n";
        color_reg_alloc.alloc_reg(functions_.back(), false);
        //std::cout << "end alloc reg:" << functions_.back()->name() << std::endl;
        functions_.back()->stack_spill_size_ = color_reg_alloc.spill_size();
        for(auto used: color_reg_alloc.used_reg) {
            if(!used.is_gp() && regs_use_fp[used.id()] == callee_save) {
                functions_.back()->callee_saved_used.emplace_back(used);
            }
        }
    }
    for (auto global : *module->get_global_variables()){
        if(global->get_data_type().is_array()){
            if(std::get<1>(global->get_init_value()) == std::vector(global->get_data_type().nr_elems(), ConstValue(0)) || (std::get<1>(global->get_init_value()) == std::vector(global->get_data_type().nr_elems(), ConstValue((float)0.0))))
                bss_globals_.push_back(new GlobalObject(global));
            else
                data_globals_.push_back(new GlobalObject(global));
        }
        else{
            if(std::get<0>(global->get_init_value()) == ConstValue(0) || (std::get<0>(global->get_init_value()) == ConstValue((float)0.0)))
                bss_globals_.push_back(new GlobalObject(global));
            else
                data_globals_.push_back(new GlobalObject(global));
        }
    }
}

void Program::gen_global_var_asm(std::ostream &out){
    if(data_globals_.size() > 0)
        out << "    .data\n";
    for (auto &data_global : data_globals_) {
        out << "    .globl " << data_global->get_name() << "\n";
        out << "    " << data_global->get_name() << ":\n";
        if(data_global->get_type().is_array()){
            for(int i = 0; i < std::get<1>(data_global->get_init_value()).size(); i++){
                if (data_global->get_type().base_type == ScalarType::Int)
                    out << "        .word " << std::fixed << std::setprecision(0) << getValue(std::get<1>(data_global->get_init_value())[i]) << "\n";
                else {
                    float val = getValue(std::get<1>(data_global->get_init_value())[i]);
                    unsigned char *bytePointer = reinterpret_cast<unsigned char*>(&val);
                    unsigned int intValue = *reinterpret_cast<unsigned int*>(bytePointer);
                    out << "        .word " << intValue << "\n";
                }
            }
        }
        else{
            if (data_global->get_type().base_type == ScalarType::Int)
                out << "        .word " << std::fixed << std::setprecision(0) << getValue(std::get<0>(data_global->get_init_value())) << "\n";
            else {
                // std::cout << data_global->get_name() << " is float\n";
                float val = getValue(std::get<0>(data_global->get_init_value()));
                unsigned char *bytePointer = reinterpret_cast<unsigned char*>(&val);
                unsigned int intValue = *reinterpret_cast<unsigned int*>(bytePointer);
                out << "        .word " << intValue << "\n";
            }
        }
    }
    if(bss_globals_.size() > 0)
        out << "    .bss\n";
    for (auto &bss_global : bss_globals_) {
        out << "    .globl " << bss_global->get_name() << "\n";
        out << "    " << bss_global->get_name() << ":\n";
        out << "        .space " << bss_global->get_type().size() << "\n";
    }
}

} // namespace riscv

} // namespace backend