#include "middleend/function_inline.hpp"
#include <iostream>

namespace middleend {

Temp *FunctionInline::new_temp(Temp *temp) {
    if(temp == nullptr)return nullptr;
    // 检查temp是否在func_arg_temp里
    for (int i = 0; i < func_arg_temp.size(); i++){
        if (func_arg_temp[i] == temp->get_index()) return call_arg_temp[i];
    }
    auto new_temp = new Temp(module_->func_map["main"]->get_temp_used() + temp->get_index(), temp->get_type());
    return new_temp;
}

void FunctionInline::new_bb(ir::BasicBlock *bb, ir::BasicBlock *new_bb, ir::BasicBlock *next_bb) {
    for (auto & inst : *bb->get_instructions()) {
        // std::cout << "inst: " << inst->to_str() << std::endl;
        ir::Instruction *new_inst = nullptr;
        if (auto assign = dynamic_cast<ir::instruction::Assign *>(inst)) {
            auto dst = new_temp(assign->getdst());
            auto src = new_temp(assign->getsrc());
            new_inst = new ir::instruction::Assign(dst, src);
        }
        if (auto li = dynamic_cast<ir::instruction::LoadImm4 *>(inst)) {
            auto dst = new_temp(li->getdst());
            new_inst = new ir::instruction::LoadImm4(dst, li->getimm());
        }
        if (auto unary = dynamic_cast<ir::instruction::Unary *>(inst)) {
            auto dst = new_temp(unary->getdst());
            auto src = new_temp(unary->getsrc());
            new_inst = new ir::instruction::Unary(unary->get_type(), dst, src);
        }
        if (auto binary = dynamic_cast<ir::instruction::Binary *>(inst)) {
            auto dst = new_temp(binary->getdst());
            auto lhs = new_temp(binary->getlhs());
            auto rhs = new_temp(binary->getrhs());
            new_inst = new ir::instruction::Binary(binary->get_type(), dst, lhs, rhs);
        }
        if (auto call = dynamic_cast<ir::instruction::Call *>(inst)) {
            auto dst = new_temp(call->getdst());
            std::vector<Temp *> args;
            for (auto src : call->get_srcs()){
                args.push_back(new_temp(src));
            }
            new_inst = new ir::instruction::Call(dst, args, call->getfunc());
        }
        if (auto branch = dynamic_cast<ir::instruction::Branch *>(inst)) {
            new_inst = new ir::instruction::Branch(bb_map[branch->get_bb()->get_index()]);
        }
        if (auto cond = dynamic_cast<ir::instruction::CondBranch *>(inst)) {
            new_inst = new ir::instruction::CondBranch(cond->get_type(), bb_map[cond->get_true_bb()->get_index()], bb_map[cond->get_false_bb()->get_index()], new_temp(cond->getcond()));
        }
        if (auto alloca = dynamic_cast<ir::instruction::Alloca *>(inst)) {
            new_inst = new ir::instruction::Alloca(new_temp(alloca->getaddr()), alloca->gettype());
        }
        if (auto store = dynamic_cast<ir::instruction::Store *>(inst)) {
            new_inst = new ir::instruction::Store(new_temp(store->getaddr()), new_temp(store->getsrc()), store->not_delete(), store->getoffset());
        }
        if (auto load = dynamic_cast<ir::instruction::Load *>(inst)) {
            new_inst = new ir::instruction::Load(new_temp(load->getdst()), new_temp(load->getaddr()), load->not_delete(), load->getoffset());
        }
        if (auto la = dynamic_cast<ir::instruction::LoadAddr *>(inst)) {
            new_inst = new ir::instruction::LoadAddr(new_temp(la->get_addr()), la->get_name());
        }
        if (auto elementptr = dynamic_cast<ir::instruction::ElementPtr *>(inst)) {
            std::vector<Temp*> indices;
            for(auto i = elementptr->get_src()->begin() + 1; i != elementptr->get_src()->end(); i++){
                indices.push_back(new_temp(*i));
            }
            new_inst = new ir::instruction::ElementPtr(new_temp(elementptr->get_dst()), new_temp(elementptr->get_base()), indices);
        }
        if (auto phi = dynamic_cast<ir::instruction::Phi *>(inst)) {
            std::vector<Temp*> values;
            std::vector<ir::BasicBlock*> bbs;
            for(auto value : *phi->getvalues()){
                values.push_back(new_temp(value));
            }
            for(auto bb : *phi->getbbs())
                bbs.push_back(bb_map[bb->get_index()]);
            new_inst = new ir::instruction::Phi(new_temp(phi->getdst()), values, bbs);
        }
        if (auto ret = dynamic_cast<ir::instruction::Return *>(inst)) {
            // 修改 return 指令为 jump
            auto value = ret->getvalue();
            if(value != nullptr){
                next_bb_bbs.push_back(new_bb);
                next_bb_values.push_back(new_temp(value));
            }
            new_inst = new ir::instruction::Branch(next_bb);
        }
        if (auto cast = dynamic_cast<ir::instruction::Cast *>(inst)) {
            auto dst = new_temp(cast->getdst());
            auto src = new_temp(cast->getsrc());
            new_inst = new ir::instruction::Cast(dst, src, cast->get_type());
        }
        new_bb->add_instruction(new_inst);
    }
}

void FunctionInline::run() {
    // 判断一个函数能否被内联
    const int LONG_FUNC_LEN = 10000;
    const int TOO_LONG_FUNC_LEN = 1000000;
    for (auto func : *module_->get_functions()) {
        if (func->is_main()) continue;
        int length = 0;
        bool flag = true;
        for (auto bb : *func->get_basic_blocks()) {
            length += bb->get_instructions()->size();
            for (auto & inst : *bb->get_instructions()) {
                if (auto call = dynamic_cast<ir::instruction::Call *>(inst)) {
                    if (call->getfunc()->get_name() == func->get_name()) {
                        flag = false;
                        break;
                    }
                }
            }
        }
        if (flag && length < LONG_FUNC_LEN) {
            can_inline_func.insert(func->get_name());
        }
    }

    // 内联
    auto main_func = module_->func_map["main"];
    bool changed = true;
    while(changed){
        changed = false;
        auto main_bbs = *main_func->get_basic_blocks();
        for (auto bb : main_bbs) {
            // std::cout << bb->get_index() << std::endl;
            for (auto inst = bb->get_instructions()->begin(); inst != bb->get_instructions()->end(); inst++){
                if (auto call = dynamic_cast<ir::instruction::Call *>(*inst)) {
                    if (can_inline_func.find(call->getfunc()->get_name()) != can_inline_func.end()) { // 进行内联
                        changed = true;
                        inline_func = call->getfunc();
                        // 新建bb
                        bb_map.clear();
                        for (auto inline_bb : *inline_func->get_basic_blocks()){
                            if(inline_bb->get_index() == 0){
                                bb_map[0] = bb;
                                continue;
                            }
                            auto new_bb_ = new ir::BasicBlock(inline_func, {}, main_func->get_bb_used() + inline_bb->get_index() - 1);
                            main_func->add_basic_block(new_bb_);
                            bb_map[inline_bb->get_index()] = new_bb_;
                        }
                        func_arg_temp.clear();
                        call_arg_temp.clear();
                        for (auto arg : call->get_srcs())
                            call_arg_temp.push_back(arg);
                        for (auto arg : *inline_func->get_arg_temp())
                            func_arg_temp.push_back(arg->get_index());
                        // 给新的bb增加指令
                        ir::BasicBlock *next_bb = nullptr;
                        if (auto branch = dynamic_cast<ir::instruction::Branch *>(*(inst + 1))) {
                            next_bb = branch->get_bb();
                        }
                        bb->get_instructions()->erase(inst, inst + 2);
                        next_bb_values.clear();
                        next_bb_bbs.clear();
                        for (auto inline_bb : *inline_func->get_basic_blocks()){
                            new_bb(inline_bb, bb_map[inline_bb->get_index()], next_bb);
                        }
                        // 给下一个bb新增assign或phi
                        if(call->getdst() != nullptr){
                            if(next_bb_values.size() == 1){
                                next_bb->add_instruction_front(new ir::instruction::Assign(call->getdst(), next_bb_values[0]));
                            }
                            else{
                                next_bb->add_instruction_front(new ir::instruction::Phi(call->getdst(), next_bb_values, next_bb_bbs));
                            }
                        }
                        main_func->set_temp_used(main_func->get_temp_used() + inline_func->get_temp_used());
                        main_func->set_bb_used(main_func->get_bb_used() + inline_func->get_bb_used());
                        // std::cout << "inline " << inline_func->get_name() << std::endl;
                        break;
                    }
                }
            }
        }
    }

}

}