#pragma once

#include "common/common.hpp"
#include "common/utils.hpp"
#include "common/type.hpp"
#include "middleend/temp.hpp"

#include <variant>

#define GetValue(x) (x.type == Int ? x.iv : x.fv)

namespace middleend{
class CFG;

namespace ir{
class Function;
class BasicBlock;

/// @brief meta for data
class DataMeta {
public:
    explicit DataMeta() = default;
    DataMeta(Type type, bool is_global, const std::string & name = "") : 
        type_(type), is_global_(is_global), name_(name) {
            if(is_global){
                if(type.is_array()){
                    if(type.base_type == Int)
                        init_value = std::vector(type.nr_elems(), ConstValue(0));
                    else init_value = std::vector(type.nr_elems(), ConstValue((float)0.0));
                }
                else{
                    if(type.base_type == Int)
                        init_value = ConstValue(0);
                    else init_value = ConstValue((float)0.0);
                }
            }
        }
    inline std::string to_str() {
        if(type_.is_array()){
            std::string ret = temp->to_str() + " : " + name_;
            for (auto dim : type_.dims) {
                if(dim == 0) ret += "[]";
                else ret += "[" + std::to_string(dim) + "]";
            }
            return ret;
        }
        else{
            return temp->to_str() + " : " + name_;
        }
    }
    inline Type get_data_type() const { return type_; }
    inline bool get_is_global() const { return is_global_; }
    inline std::string get_name() const { return name_; }
    inline Temp *get_temp() const { return temp; }
    inline void set_temp(Temp *temp) { this->temp = temp; }
    inline void set_init_value(ConstValue init_value) {
        if (type_.base_type == Int)
            this->init_value = (int)(init_value.type == Int ? init_value.iv : init_value.fv);
        else if (type_.base_type == Float)
            this->init_value = (float)(init_value.type == Int ? init_value.iv : init_value.fv);
    }
    inline void set_init_value(std::vector<ConstValue> init_value) { this->init_value = init_value; }
    inline void add_init_value(ConstValue init_value) { std::get<1>(this->init_value).push_back(init_value); }
    inline std::variant<ConstValue, std::vector<ConstValue>> get_init_value() const { return init_value; }
protected:
    Type type_;
    bool is_global_;
    std::string name_;
    Temp *temp;
    std::variant<ConstValue, std::vector<ConstValue>> init_value;
};

/// @brief variable
class DataVar : public DataMeta {
public:
    explicit DataVar() = default;
    DataVar(Type var_type, bool is_global, const std::string & name = "") : 
        DataMeta(var_type, is_global, name) {}
    // inline std::string to_str() { return temp->to_str() + " = " + name_; }
};

// array
class DataArray : public DataMeta {
public:
    explicit DataArray() = default;
    DataArray(Type var_type, bool is_global, const std::string & name = "") : 
        DataMeta(var_type, is_global, name) {}
    // inline std::string to_str() {
    //     std::string ret = temp->to_str() + " = " + name_;
    //     for (auto dim : type_.dims) {
    //         ret += "[" + std::to_string(dim) + "]";
    //     }
    //     return ret;
    // }
};

enum IRValueType {
    Value_Argument,
    Value_BasicBlock,
    Value_Constant,
    Value_Function,
    Value_GlobalVariable,
    Value_Instruction
};

/// @brief a typed value that may be used (among other things) as an operand to an instruction
class Value {
public:
    explicit Value() = default;
    Value(IRValueType type, const std::string & name = "") : type_(type), name_(name) {}
    std::string to_str() { return "none to_str method for Value(" + name_ + ")"; }
    IRValueType get_type() const { return type_; }
    std::string get_name() const { return name_; }
protected:
    IRValueType type_;
    std::string name_;
};

enum IRInstrKind {
    IRInstr_LABEL,
    IRInstr_SEQ,
    IRInstr_JMP,
    IRInstr_JMP_COND,
    IRInstr_CALL,
    IRInstr_RET,
};

/// @brief the common base class for instructions
class Instruction : public Value {
public:
    explicit Instruction() = default;
    Instruction(IRInstrKind instrkind, const std::vector<Temp*> & dsts, const std::vector<Temp*> & srcs) : 
        Value(IRValueType::Value_Instruction), instrkind_(instrkind), dsts_(dsts), srcs_(srcs), parent_(nullptr) {}
    virtual std::string to_str() = 0;
    inline void set_parent(BasicBlock* parent) { parent_ = parent; }
    inline ir::BasicBlock* get_parent() const { return parent_; }
    inline IRInstrKind get_instr_kind() const { return instrkind_; }
    inline std::vector<Temp*>* get_dst() { return &dsts_; }
    inline std::vector<Temp*>* get_src() { return &srcs_; }
    void print(std::ostream &out) {
        if(instrkind_ != IRInstr_LABEL)print_indent(out, INDENT_LEN);
        out << to_str() << "\n";
    }
protected:
    BasicBlock* parent_;
    IRInstrKind instrkind_;
    std::vector<Temp*> dsts_;
    std::vector<Temp*> srcs_;
};

/// @brief a single entry single exit section of the code
class BasicBlock : public Value {
public:
    explicit BasicBlock() = default;
    BasicBlock(Function* parent, const std::vector<Instruction*> &instructions, int index) : 
        Value(IRValueType::Value_BasicBlock, "_B" + std::to_string(index)), parent_(parent), instructions_(instructions), index_(index) {}
    inline std::vector<Instruction*>* get_instructions() { return &instructions_; }
    inline void add_instruction(Instruction* ins) { instructions_.push_back(ins); }
    inline void add_instruction_front(Instruction* ins) { instructions_.insert(instructions_.begin(), ins); }
    void print(std::ostream &out) const {
        out << name_ << ":\n";
        for (auto &child : instructions_) {
            child->print(out);
        }
    }
    inline int get_index() const { return index_; }
protected:
    int index_;
    Function* parent_;
    std::vector<Instruction*> instructions_;
    std::vector<Temp*> define;
    std::vector<Temp*> live_in;
    std::vector<Temp*> live_out;
    std::vector<Temp*> live_use;
};

class Module;

/// @brief a single procedure in the program
class Function : public Value {
public:
    explicit Function() = default;
    Function(Module *parent, Type ret_type, const std::vector<DataMeta*> &arguments, const std::string & name) : 
        Value(IRValueType::Value_Function, name), ret_type_(ret_type), parent_(parent), arguments_(arguments), basicblock_used_(0) {}
    inline bool is_main() { return name_ == "main"; }
    inline bool has_return_value() { return ret_type_.base_type != Void; }
    inline Type get_ret_type() const { return ret_type_; }
    inline Module * get_parent() const { return parent_; }
    inline int get_arg_num() const { return arguments_.size(); }
    inline const std::vector<DataMeta*>* get_arguments() const { return &arguments_; }
    inline std::vector<BasicBlock*>* get_basic_blocks() { return &basic_blocks_; }
    inline const std::vector<Instruction*>* get_instructions() const { return &instructions_; }
    inline void add_argument(DataMeta *arg) { arguments_.push_back(arg); }
    inline void add_instruction(Instruction *inst) { instructions_.push_back(inst); }
    inline void add_basic_block(BasicBlock *bb) { basic_blocks_.push_back(bb); }
    inline BasicBlock* get_entry() const { return basic_blocks_[0]; }
    int get_temp_used() { return temp_used_; }
    void set_temp_used(int used) { temp_used_ = used; }
    int get_bb_used() { return basicblock_used_; }
    void set_bb_used(int used) { basicblock_used_ = used; }
    void add_arg_temp(Temp *temp){ arg_temp.push_back(temp); }
    std::vector<Temp*>* get_arg_temp() {return &arg_temp;}
    void print(std::ostream &out) const {
        out << type_string(ret_type_) << ' ' << name_ << "(";
        for (int i = 0; i < arg_temp.size(); ++i) {
            out << arg_temp[i]->to_str();
            if(i != arg_temp.size() - 1)
                out << ", ";
        }
        out << ") {\n";
        for (auto &child : basic_blocks_) {
            child->print(out);
        }
        out << "}\n";
    }
protected:
    Type ret_type_;
    Module *parent_;
    std::vector<DataMeta*> arguments_;
    std::vector<Temp*> arg_temp;
    std::vector<BasicBlock*> basic_blocks_;
    std::vector<Instruction*> instructions_;
    int temp_used_;
    int basicblock_used_;
};

class LibFunction: public Function{
public:
    explicit LibFunction() = default;
    LibFunction(Module *parent, Type ret_type, const std::vector<DataMeta*> &arguments, const std::string & name) : 
        Function(parent, ret_type, arguments, name) {}
};

namespace instruction{
class Assign : public Instruction {
public:
    explicit Assign() = default;
    Assign(Temp* dst, Temp* src) : 
        Instruction(IRInstrKind::IRInstr_SEQ, std::vector<Temp*>{dst}, std::vector<Temp*>{src}) {}
    inline std::string to_str() { return dsts_[0]->to_str() + " = " + srcs_[0]->to_str(); }
    inline Temp* getdst() const { return dsts_[0]; }
    inline Temp* getsrc() const { return srcs_[0]; }
};

class LoadImm4 : public Instruction {
public:
    explicit LoadImm4() = default;
    LoadImm4(Temp* dst, ConstValue imm) : 
        Instruction(IRInstrKind::IRInstr_SEQ, std::vector<Temp*>{dst}, std::vector<Temp*>{}), imm_(imm) {}
    inline std::string to_str() { return dsts_[0]->to_str() + " = " + imm_.to_string(); }
    inline Temp* getdst() const { return dsts_[0]; }
    inline ConstValue getimm() const { return imm_; }
protected:
    ConstValue imm_;
};

class Unary : public Instruction {
public:
    // enum IRInstrUnary {
    //     IR_Instr_NEG,
    //     IR_Instr_BITNOT,
    //     IR_Instr_LOGICNOT
    // };
    explicit Unary() = default;
    Unary(UnaryOp type, Temp* dst, Temp* src) : 
        Instruction(IRInstrKind::IRInstr_SEQ, std::vector<Temp*>{dst}, std::vector<Temp*>{src}), type_(type) {}
    inline std::string to_str() { 
        if (type_ == UnaryOp::Add) return dsts_[0]->to_str() + " = +" + srcs_[0]->to_str();
        if (type_ == UnaryOp::Sub) return dsts_[0]->to_str() + " = -" + srcs_[0]->to_str();
        if (type_ == UnaryOp::Not) return dsts_[0]->to_str() + " = !" + srcs_[0]->to_str();
        assert(false);
    }
    inline ConstValue to_const(ConstValue src) {
        if (type_ == UnaryOp::Add) return src;
        if (type_ == UnaryOp::Sub) return src.getNeg();
        if (type_ == UnaryOp::Not) return src.getNot();
        assert(false);
    }
    inline Temp* getdst() const { return dsts_[0]; }
    inline Temp* getsrc() const { return srcs_[0]; }
    inline UnaryOp get_type() const { return type_; }
protected:
    UnaryOp type_;
};

class Binary : public Instruction {
public:
    explicit Binary() = default;
    Binary(BinaryOp type, Temp* dst, Temp* lhs, Temp* rhs) : 
        Instruction(IRInstrKind::IRInstr_SEQ, std::vector<Temp*>{dst}, std::vector<Temp*>{lhs, rhs}), type_(type){}
    inline std::string to_str() { 
        if (type_ == BinaryOp::And) return dsts_[0]->to_str() + " = " + srcs_[0]->to_str() + " & " + srcs_[1]->to_str();
        if (type_ == BinaryOp::Or ) return dsts_[0]->to_str() + " = " + srcs_[0]->to_str() + " | " + srcs_[1]->to_str();

        if (type_ == BinaryOp::Add) return dsts_[0]->to_str() + " = " + srcs_[0]->to_str() + " + " + srcs_[1]->to_str();
        if (type_ == BinaryOp::Sub) return dsts_[0]->to_str() + " = " + srcs_[0]->to_str() + " - " + srcs_[1]->to_str();
        if (type_ == BinaryOp::Mul) return dsts_[0]->to_str() + " = " + srcs_[0]->to_str() + " * " + srcs_[1]->to_str();
        if (type_ == BinaryOp::Div) return dsts_[0]->to_str() + " = " + srcs_[0]->to_str() + " / " + srcs_[1]->to_str();
        if (type_ == BinaryOp::Mod) return dsts_[0]->to_str() + " = " + srcs_[0]->to_str() + " % " + srcs_[1]->to_str();
        
        if (type_ == BinaryOp::Eq) return dsts_[0]->to_str() + " = " + srcs_[0]->to_str() + " == " + srcs_[1]->to_str();
        if (type_ == BinaryOp::Neq) return dsts_[0]->to_str() + " = " + srcs_[0]->to_str() + " != " + srcs_[1]->to_str();
        if (type_ == BinaryOp::Lt) return dsts_[0]->to_str() + " = " + srcs_[0]->to_str() + " < " + srcs_[1]->to_str();
        if (type_ == BinaryOp::Leq) return dsts_[0]->to_str() + " = " + srcs_[0]->to_str() + " <= " + srcs_[1]->to_str();
        if (type_ == BinaryOp::Gt) return dsts_[0]->to_str() + " = " + srcs_[0]->to_str() + " > " + srcs_[1]->to_str();
        if (type_ == BinaryOp::Geq) return dsts_[0]->to_str() + " = " + srcs_[0]->to_str() + " >= " + srcs_[1]->to_str();
        if (type_ == BinaryOp::Shr) return dsts_[0]->to_str() + " = " + srcs_[0]->to_str() + " >> " + srcs_[1]->to_str();
        if (type_ == BinaryOp::Shl) return dsts_[0]->to_str() + " = " + srcs_[0]->to_str() + " << " + srcs_[1]->to_str();
        return "error";
    }
    inline ConstValue to_const(ConstValue lhs, ConstValue rhs) {
        if (type_ == BinaryOp::And) return ConstValue((int)GetValue(lhs) & (int)GetValue(rhs));
        if (type_ == BinaryOp::Or ) return ConstValue((int)GetValue(lhs) | (int)GetValue(rhs));

        if (lhs.type == Float || rhs.type == Float) {
            if (type_ == BinaryOp::Add) return ConstValue(GetValue(lhs) + GetValue(rhs));
            if (type_ == BinaryOp::Sub) return ConstValue(GetValue(lhs) - GetValue(rhs));
            if (type_ == BinaryOp::Mul) return ConstValue(GetValue(lhs) * GetValue(rhs));
            if (type_ == BinaryOp::Div) return ConstValue(GetValue(lhs) / GetValue(rhs));
        }
        else {
            if (type_ == BinaryOp::Add) return ConstValue((int)GetValue(lhs) + (int)GetValue(rhs));
            if (type_ == BinaryOp::Sub) return ConstValue((int)GetValue(lhs) - (int)GetValue(rhs));
            if (type_ == BinaryOp::Mul) return ConstValue((int)GetValue(lhs) * (int)GetValue(rhs));
            if (type_ == BinaryOp::Div) return ConstValue((int)GetValue(lhs) / (int)GetValue(rhs));
        }
        if (type_ == BinaryOp::Mod) return ConstValue((int)GetValue(lhs) % (int)GetValue(rhs));
        
        if (type_ == BinaryOp::Eq) return lhs == rhs;
        if (type_ == BinaryOp::Neq) return lhs != rhs;

        if (type_ == BinaryOp::Lt) return ConstValue(GetValue(lhs) < GetValue(rhs));
        if (type_ == BinaryOp::Leq) return ConstValue(GetValue(lhs) <= GetValue(rhs));
        if (type_ == BinaryOp::Gt) return ConstValue(GetValue(lhs) > GetValue(rhs));
        if (type_ == BinaryOp::Geq) return ConstValue(GetValue(lhs) >= GetValue(rhs));
        if (type_ == BinaryOp::Shr) return ConstValue((int)GetValue(lhs) >> (int)GetValue(rhs));
        if (type_ == BinaryOp::Shl) return ConstValue((int)GetValue(lhs) << (int)GetValue(rhs));
        assert(false);
    }
    inline Temp* getdst() const { return dsts_[0]; }
    inline Temp* getlhs() const { return srcs_[0]; }
    inline Temp* getrhs() const { return srcs_[1]; }
    inline BinaryOp get_type() const { return type_; }
protected:
    BinaryOp type_;
};

class Label : public Instruction {
public:
    explicit Label() = default;
    Label(const std::string & label) : 
        Instruction(IRInstrKind::IRInstr_LABEL, std::vector<Temp*>{}, std::vector<Temp*>{}), label_(label) {}
    inline std::string to_str() { return label_; }
    inline std::string get_label() const { return label_; }
    inline void set_label(const std::string & label) { label_ = label; }
protected:
    std::string label_;
};

class Call : public Instruction {
public:
    explicit Call() = default;
    Call(Temp* dst, const std::vector<Temp*> & srcs, Function* func) : 
        Instruction(IRInstrKind::IRInstr_CALL, std::vector<Temp*>{dst}, srcs), func_(func) {}
    inline Temp* getdst() const { return dsts_[0]; }
    inline Function* getfunc() const { return func_; }
    inline std::vector<Temp *> get_srcs() const { return srcs_; }
    inline std::string to_str() { 
        std::string ret = "";
        if(dsts_[0] != nullptr)ret = dsts_[0]->to_str() + " = ";
        ret += "call " + func_->get_name() + "(";
        for (int i = 0; i < srcs_.size(); ++i) {
            ret += srcs_[i]->to_str();
            if (i != srcs_.size() - 1) ret += ", ";
        }
        ret += ")";
        return ret;
    }
protected:
    Function* func_;
};

class Branch : public Instruction {
public:
    explicit Branch() = default;
    Branch(BasicBlock* bb) : 
        Instruction(IRInstrKind::IRInstr_JMP, std::vector<Temp*>{}, std::vector<Temp*>{}), bb_(bb) {}
    inline std::string to_str() { return "jump " + bb_->get_name();  }
    inline BasicBlock* get_bb() const { return bb_; }
protected:
    BasicBlock* bb_;
};

class CondBranch : public Instruction {
public:
    enum IRInstrCondBranch {
        IR_Instr_BEQ,
        IR_Instr_BNE
    };
    explicit CondBranch() = default;
    CondBranch(IRInstrCondBranch type, BasicBlock* true_bb, BasicBlock* false_bb, Temp* cond) : 
        Instruction(IRInstrKind::IRInstr_JMP_COND, std::vector<Temp*>{}, std::vector<Temp*>{cond}), type_(type), true_bb_(true_bb), false_bb_(false_bb) {}
    inline std::string to_str() { return "if " + srcs_[0]->to_str() + " " + (type_ == IR_Instr_BEQ ? "==" : "!=") + " 0 jump " + false_bb_->get_name() + " else jump " + true_bb_->get_name(); }
    inline Temp* getcond() const { return srcs_[0]; }
    inline BasicBlock* get_true_bb() const { return true_bb_; }
    inline BasicBlock* get_false_bb() const { return false_bb_; }
    inline IRInstrCondBranch get_type() const { return type_; }
protected:
    IRInstrCondBranch type_;
    BasicBlock* true_bb_;
    BasicBlock* false_bb_;
};

class Return : public Instruction {
public:
    explicit Return() = default;
    Return(Temp* value) : 
        Instruction(IRInstrKind::IRInstr_RET, std::vector<Temp*>{}, std::vector<Temp*>{value}) {}
    inline std::string to_str() {
        if(srcs_[0] == nullptr)return "return";
        return "return " + srcs_[0]->to_str();
    }
    inline bool has_return_value() const { return srcs_[0] != nullptr; }
    inline Temp* getvalue() const { return srcs_[0]; }
};

class Alloca : public Instruction {
public:
    explicit Alloca() = default;
    Alloca(Temp* addr, Type type) : 
        Instruction(IRInstrKind::IRInstr_SEQ, std::vector<Temp*>{addr}, std::vector<Temp*>{}), type_(type) {}
    inline std::string to_str() { return "alloca " + dsts_[0]->to_str() + " = " + std::to_string(type_.size()); }
    inline int get_size() { return type_.size(); }
    inline Temp* getaddr() const { return dsts_[0]; }
    inline Type gettype() const { return type_; }
protected:
    Type type_;
};

class Store : public Instruction {
public:
    explicit Store() = default;
    Store(Temp* addr, Temp* src, bool not_delete, int offset = 0) : 
        Instruction(IRInstrKind::IRInstr_SEQ, std::vector<Temp*>{addr}, std::vector<Temp*>{src}), offset_(offset), not_delete_(not_delete) {}
    inline std::string to_str() { return "store *(" + dsts_[0]->to_str() + " + " + std::to_string(offset_) + ") = " + srcs_[0]->to_str(); }
    inline Temp* getsrc() const { return srcs_[0]; }
    inline Temp* getaddr() const { return dsts_[0]; }
    inline int getoffset() const { return offset_; }
    inline bool not_delete() const { return not_delete_; }
protected:
    int offset_;
    bool not_delete_;
};

class Load : public Instruction {
public:
    explicit Load() = default;
    Load(Temp* dst, Temp* addr, bool not_delete, int offset = 0) : 
        Instruction(IRInstrKind::IRInstr_SEQ, std::vector<Temp*>{dst}, std::vector<Temp*>{addr}), offset_(offset), not_delete_(not_delete) {}
    inline std::string to_str() { return "load " + dsts_[0]->to_str() + " = *" + srcs_[0]->to_str(); }
    inline Temp* getdst() const { return dsts_[0]; }
    inline Temp* getaddr() const { return srcs_[0]; }
    inline int getoffset() const { return offset_; }
    inline bool not_delete() const { return not_delete_; }
protected:
    int offset_;
    bool not_delete_;
};

class LoadAddr : public Instruction {
public:
    explicit LoadAddr() = default;
    LoadAddr(Temp* addr, std::string name) : 
        Instruction(IRInstrKind::IRInstr_SEQ, std::vector<Temp*>{addr}, std::vector<Temp*>{}), name_(name) {}
    inline std::string to_str() { return dsts_[0]->to_str() + " = &" + name_; }
    inline Temp* get_addr() const {return dsts_[0]; }
    inline std::string get_name() const {return name_; }
protected:
    std::string name_;
};

class ElementPtr : public Instruction {
public:
    explicit ElementPtr() = default;
    ElementPtr(Temp* dst, Temp* base, std::vector<Temp*> indices) : 
        Instruction(IRInstrKind::IRInstr_SEQ, std::vector<Temp*>{dst}, std::vector<Temp*>{}) {
            srcs_.push_back(base);
            srcs_.insert(srcs_.end(), indices.begin(), indices.end());
        }  
    inline std::string to_str() {
        std::string s = dsts_[0]->to_str() + " = elementptr: " + srcs_[0]->to_str();
        auto indices_ = std::vector<Temp*>(srcs_.begin() + 1, srcs_.end());
        for (auto i : indices_) {
            s += "[" + i->to_str() + "]";
        }
        return s;
    }
    inline Temp* get_dst() const { return dsts_[0]; }
    inline Temp* get_base() const { return srcs_[0]; }
protected:
};

class Phi : public Instruction {
public:
    explicit Phi() = default;
    Phi(Temp* dst, const std::vector<Temp*> & values, const std::vector<BasicBlock*> & bbs) : 
        Instruction(IRInstrKind::IRInstr_SEQ, std::vector<Temp*>{dst}, values), bbs_(bbs) {}
    inline std::string to_str() { 
        std::string ret = dsts_[0]->to_str() + " = phi ";
        for (int i = 0; i < srcs_.size(); ++i) {
            ret += "[" + srcs_[i]->to_str() + ", " + bbs_[i]->get_name() + "]";
            if (i != srcs_.size() - 1) ret += ", ";
        }
        return ret;
    }
    inline int getsize() const { return srcs_.size(); }
    inline Temp* getdst() const { return dsts_[0]; }
    inline void setdst(Temp *temp) { dsts_[0] = temp;}
    inline std::vector<Temp*>* getvalues() { return &srcs_; }
    inline std::vector<BasicBlock*>* getbbs() { return &bbs_; }
protected:
    std::vector<BasicBlock*> bbs_;
};

class Cast : public Instruction {
public:
    explicit Cast() = default;
    Cast(Temp* dst, Temp* src, Type type) : 
        Instruction(IRInstrKind::IRInstr_SEQ, std::vector<Temp*>{dst}, std::vector<Temp*>{src}), type_(type) {}
    inline std::string to_str() { return "cast " + dsts_[0]->to_str() + " = " + srcs_[0]->to_str(); }
    inline Temp* getdst() const { return dsts_[0]; }
    inline Temp* getsrc() const { return srcs_[0]; }
    inline Type gettype() const { return type_; }
protected:
    Type type_;
};
} // namespace instruction

/// @brief functions, global variables and a symboltable
class Module {
public:
    explicit Module() = default;
    Module(const std::vector<Function*> & functions, const std::vector<DataMeta*> global_variables) : 
        functions_(functions), global_variables_(global_variables) {
            if (!check_global_variables()) {
                printf("global variables are not global.\n");
                assert(false);
            }
        }
    inline const std::vector<Function*> * get_functions() const { return &functions_; }
    inline const std::vector<DataMeta*> * get_global_variables() const { return &global_variables_; }
    inline std::unordered_map<std::string, Function*> get_func_map() { return func_map; }
    inline void add_function(Function * function) {
        // func_map.emplace(function->get_name(), function);
        functions_.push_back(function);
    }
    inline void add_global_variable(DataMeta * global_variable) { global_variables_.push_back(global_variable); }
    bool check_global_variables() {
        for (auto global_variable : global_variables_) {
            if (!global_variable->get_is_global()) {
                return false;
            }
        }
        return true;
    }
    void print(std::ostream &out) const {
        for (auto &global : global_variables_){
            if(global->get_data_type().is_array()){
                out << "init global " << type_string(global->get_data_type()) << ' ' << global->get_name() << std::endl;
            }
            else{
                if(std::get<0>(global->get_init_value()).type == Int)
                    out << "init global " << type_string(global->get_data_type()) << ' ' << global->get_name() << " with " << std::get<0>(global->get_init_value()).iv << std::endl;
                else{
                    out << "init global " << type_string(global->get_data_type()) << ' ' << global->get_name() << " with " << std::get<0>(global->get_init_value()).fv << std::endl;
                }
            }
        }
        for (auto &child : functions_) {
            child->print(out);
        }
    }
    virtual ~Module() {};
    std::unordered_map<std::string, Function*> func_map;
    std::unordered_map<std::string, LibFunction*> lib_funcs;
protected:
    std::vector<Function*> functions_;
    std::vector<DataMeta*> global_variables_;
};

/// @brief Valid data type : var, varptr, array, arrayptr
// enum DataType {
//     Var,
//     VarPtr,
//     Array,
//     ArrayPtr
// };

/// @brief variable type
// enum VarType {
//     Int,
//     Float,
//     Double,
//     Char,
//     Bool,
//     Void
// };


/// @brief a connection between Value and User
class Use {

};

/// @brief the common base class of IR that may refer to Values
class User : public Value {
protected:
    std::vector<Value*> operands_;
public:
    explicit User() = default;
    User(IRValueType type, const std::vector<Value*> & operands, const std::string & name = "") : Value(type, name), operands_(operands) {}
};

} // namespace ir

} // namespace middleend